/*
 * Copyright (C) 2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file select.c \brief The select mode of adu. */

#include <dirent.h> /* readdir() */
#include <sys/types.h>
#include <regex.h>

#include "format.h"
#include "adu.h"
#include "gcc-compat.h"
#include "cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "user.h"
#include "select.cmdline.h"

/** \cond */
/* global list */
#define GLOBAL_LIST_ATOMS \
	ATOM(size, SIZE) \
	ATOM(files, COUNT) \
	ATOM(dirname, STRING) \

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom global_list_atoms[] = {
	GLOBAL_LIST_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) gla_ ## x,
enum global_list_atoms {GLOBAL_LIST_ATOMS};
#undef ATOM

/* global summary */
#define GLOBAL_SUMMARY_ATOMS \
	ATOM(dirs, COUNT) \
	ATOM(files, COUNT) \
	ATOM(size, SIZE)

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom global_summary_atoms[] = {
	GLOBAL_SUMMARY_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) gsa_ ## x,
enum global_summary_atoms {GLOBAL_SUMMARY_ATOMS};
#undef ATOM

/* user list */
#define USER_LIST_ATOMS \
	ATOM(pw_name, STRING) \
	ATOM(uid, ID) \
	ATOM(size, SIZE) \
	ATOM(files, COUNT) \
	ATOM(dirname, STRING) \

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom user_list_atoms[] = {
	USER_LIST_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) ula_ ## x,
enum user_list_atoms {USER_LIST_ATOMS};
#undef ATOM

/* user list header */
#define USER_LIST_HEADER_TRAILER_ATOMS \
	ATOM(pw_name, STRING) \
	ATOM(uid, ID)

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom user_list_header_trailer_atoms[] = {
	USER_LIST_HEADER_TRAILER_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) ulha_ ## x,
enum user_list_header_trailer_atoms {USER_LIST_HEADER_TRAILER_ATOMS};
#undef ATOM

/* user summary */
#define USER_SUMMARY_ATOMS \
	ATOM(pw_name, STRING) \
	ATOM(uid, ID) \
	ATOM(dirs, COUNT) \
	ATOM(files, COUNT) \
	ATOM(size, SIZE)

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom user_summary_atoms[] = {
	USER_SUMMARY_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) usa_ ## x,
enum user_summary_atoms {USER_SUMMARY_ATOMS};
#undef ATOM

/** \endcond */

struct global_list_info {
	uint32_t count;
	int ret;
	int osl_errno;
	struct format_info *fi;
	regex_t *preg;
	int inverse_matching;
};

struct global_summary_info {
	/** Global dir count. */
	uint64_t num_dirs;
	/** Global files count. */
	uint64_t num_files;
	/** Global bytes count. */
	uint64_t num_bytes;
	regex_t *preg;
	int inverse_matching;
	int ret;
	int osl_errno;
};

struct user_list_info {
	uint32_t count;
	struct user_info *ui;
	struct format_info *fi;
	regex_t *preg;
	int inverse_matching;
	int ret;
	int osl_errno;
};

struct user_list_format_info {
	struct format_info *fi;
	struct format_info *header_fi;
	struct format_info *trailer_fi;
};

struct user_summary_info {
	struct user_info *ui;
	/** Total number of files owned by this user. */
	uint64_t files;
	/** Total number of bytes owned by this user. */
	uint64_t bytes;
	/** Total number of directories that contain at least one file */
	uint64_t dirs;
	int ret;
	int osl_errno;
	regex_t *preg;
	int inverse_matching;
};

struct user_summary_loop_data {
	unsigned num_admissible_users;
	struct user_summary_info *usis;
	struct user_summary_info *current;
	struct format_info *fi;
};

static FILE *output_file;

__printf_1_2 static int output(const char const *fmt, ...)
{
	va_list argp;
	int ret;

	va_start(argp, fmt);
	ret = vfprintf(output_file, fmt, argp);
	va_end(argp);
	return ret < 0? -E_OUTPUT : 1;
}

static int get_dir_name_by_number(uint64_t *dirnum, char **name)
{
	char *result = NULL, *tmp;
	struct osl_row *row;
	uint64_t val = *dirnum;
	struct osl_object obj;
	int ret;
	char *pfx;

again:
	obj.data = &val;
	obj.size = sizeof(val);
	ret = osl(osl_get_row(dir_table, DT_NUM, &obj, &row));
	if (ret < 0)
		goto out;
	ret = osl(osl_get_object(dir_table, row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	val = *(uint64_t *)obj.data;
	ret = osl(osl_get_object(dir_table, row, DT_NAME, &obj));
	if (ret < 0)
		goto out;
	pfx = (select_conf.print_base_dir_given || val)? (char *)obj.data :  ".";
	tmp = make_message("%s/%s", pfx, result? result : "");
	free(result);
	result = tmp;
	if (val)
		goto again;
out:
	if (ret < 0) {
		free(result);
		*name = NULL;
	} else {
		assert(result);
		*name = result;
	}
	return ret;
}

static int get_dir_name_of_row(struct osl_row *dir_table_row, char **name)
{
	struct osl_object obj;
	int ret;

	*name = NULL;
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_NUM, &obj));
	if (ret < 0)
		return ret;
	return get_dir_name_by_number((uint64_t *)obj.data, name);
}

static int get_dir_name_of_user_row(struct osl_row *user_table_row,
		struct user_info *ui, char **dirname)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, user_table_row,
		UT_DIR_NUM, &obj));

	if (ret < 0)
		return ret;
	return get_dir_name_by_number((uint64_t *)obj.data, dirname);
}

static int get_num_files_of_row(struct osl_row *row, uint64_t *num_files)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(dir_table, row, DT_FILES, &obj));
	if (ret < 0)
		return ret;
	*num_files = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_user_files(struct osl_row *row, struct user_info *ui,
		uint64_t *num_files)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, row, UT_FILES, &obj));

	if (ret < 0)
		return ret;
	*num_files = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_bytes_of_row(struct osl_row *row, uint64_t *num_bytes)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(dir_table, row, DT_BYTES, &obj));
	if (ret < 0)
		return ret;
	*num_bytes = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_user_bytes(struct osl_row *row, struct user_info *ui,
		uint64_t *num_bytes)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, row, UT_BYTES, &obj));

	if (ret < 0)
		return ret;
	*num_bytes = *(uint64_t *)obj.data;
	return 1;
}

static void free_regex(regex_t *preg)
{
	if (!preg)
		return;
	regfree(preg);
	free(preg);
}

static int compile_regex(regex_t **preg, int *invert)
{
	int ret;
	size_t size;
	char *buf, *p = select_conf.pattern_arg;

	if (!select_conf.pattern_given || !p[0]) {
		*preg = NULL;
		return 0;
	}
	if (p[0] == '!') {
		if (!p[1]) {
			*preg = NULL;
			return -E_REGEX;
		}
		*invert = 1;
		p++;
	} else
		*invert = 0;
	*preg = adu_malloc(sizeof(regex_t));
	ret = regcomp(*preg, p, 0);
	if (!ret)
		return 1;
	size = regerror(ret, *preg, NULL, 0);
	buf = adu_malloc(size);
	regerror(ret, *preg, buf, size);
	ERROR_LOG("%s\n", buf);
	free(buf);
	free_regex(*preg);
	*preg = NULL;
	return -E_REGEX;
}

static int dir_is_admissible(char *dirname, regex_t *preg, int inverse_matching)
{
	int ret;

	if (!preg)
		return 1;
	ret = regexec(preg, dirname, 0, NULL, 0);
	if (ret == REG_NOMATCH && !inverse_matching)
		return 0;
	if (ret != REG_NOMATCH && inverse_matching)
		return 0;
	return 1;
}

static int check_loop_return(int ret, int loop_ret, int loop_osl_errno)
{
	if (ret >= 0)
		return ret;
	assert(ret == -E_OSL);
	if (osl_errno != E_OSL_LOOP)
		/* error not caused by loop function returning negative. */
		return ret;
	assert(loop_ret < 0);
	if (loop_ret == -E_LOOP_COMPLETE) /* no error */
		return 1;
	if (loop_ret == -E_OSL) { /* osl error in loop function */
		assert(loop_osl_errno);
		osl_errno = loop_osl_errno;
	}
	return loop_ret;
}

static int adu_loop_reverse(struct osl_table *t, unsigned col_num, void *private_data,
		osl_rbtree_loop_func *func, int *loop_ret, int *loop_osl_errno)
{
	int ret = osl(osl_rbtree_loop_reverse(t, col_num, private_data, func));
	return check_loop_return(ret, *loop_ret, *loop_osl_errno);
}

static int global_summary_loop_function(struct osl_row *row, void *data)
{
	struct global_summary_info *gsi = data;
	int ret;
	uint64_t num;

	if (gsi->preg) {
		char *dirname;
		ret = get_dir_name_of_row(row, &dirname);
		if (ret < 0)
			goto err;
		ret = dir_is_admissible(dirname, gsi->preg, gsi->inverse_matching);
		free(dirname);
		if (!ret)
			return 1;
	}

	ret = get_num_files_of_row(row, &num);
	if (ret < 0)
		goto err;
	gsi->num_files += num;

	ret = get_num_bytes_of_row(row, &num);
	if (ret < 0)
		goto err;
	gsi->num_bytes += num;
	gsi->num_dirs++;
	return 1;
err:
	gsi->ret = ret;
	gsi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int print_global_summary(struct format_info *fi)
{
	int ret;
	char *buf;
	struct global_summary_info gsi = {.num_dirs = 0};
	char *header = select_conf.header_given? select_conf.header_arg :
		"Global summary\n";

	union atom_value values[] = {
		[gsa_dirs] = {.num_value = 0ULL},
		[gsa_files] = {.num_value =  0ULL},
		[gsa_size] = {.num_value =  0ULL}
	};

	ret = compile_regex(&gsi.preg, &gsi.inverse_matching);
	if (ret < 0)
		return ret;
	ret = adu_loop_reverse(dir_table, DT_BYTES, &gsi,
		global_summary_loop_function, &gsi.ret, &gsi.osl_errno);
	free_regex(gsi.preg);
	if (ret < 0)
		return ret;
	values[gsa_dirs].num_value = (long long unsigned)gsi.num_dirs;
	values[gsa_files].num_value = (long long unsigned)gsi.num_files;
	values[gsa_size].num_value = (long long unsigned)gsi.num_bytes;

	ret = output("%s", header);
	if (ret < 0)
		return ret;
	buf = format_items(fi, values);
	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		return ret;
	return output("%s", select_conf.trailer_arg);
}

/* row: a pointer to a row of the *user* table */
static int user_summary_loop_function(struct osl_row *row, void *data)
{
	struct user_summary_info *usi = data;
	uint64_t num;
	int ret;

	if (usi->preg) {
		char *dirname;
		ret = get_dir_name_of_user_row(row, usi->ui, &dirname);
		if (ret < 0)
			goto err;
		ret = dir_is_admissible(dirname, usi->preg, usi->inverse_matching);
		free(dirname);
		if (!ret)
			return 1;
	}
	ret = get_num_user_files(row, usi->ui, &num);
	if (ret < 0)
		goto err;
	usi->files += num;
	ret = get_num_user_bytes(row, usi->ui, &num);
	if (ret < 0)
		goto err;
	usi->bytes += num;
	usi->dirs++;
	return 1;
err:
	usi->ret = ret;
	usi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int compute_user_summary(struct user_info *ui, void *data)
{
	struct user_summary_loop_data *usld = data;
	struct user_summary_info *usi = usld->current++;
	int ret = compile_regex(&usi->preg, &usi->inverse_matching);

	if (ret < 0)
		return ret;
	usi->ui = ui;
	ret = adu_loop_reverse(ui->table, UT_BYTES, usi, user_summary_loop_function,
		&usi->ret, &usi->osl_errno);
	free_regex(usi->preg);
	return ret;
}

static int print_user_summary_line(struct user_summary_info *usi,
		struct format_info *fi)
{
	struct user_info *ui = usi->ui;
	union atom_value values[] = {
		[usa_pw_name] = {.string_value = ui->pw_name?
			ui->pw_name : "?"},
		[usa_uid] = {.num_value = (long long unsigned)ui->uid},
		[usa_dirs] = {.num_value = (long long unsigned)usi->dirs},
		[usa_files] = {.num_value = (long long unsigned)usi->files},
		[usa_size] = {.num_value =  (long long unsigned)usi->bytes}
	};
	char *buf = format_items(fi, values);
	int ret = output("%s", buf);

	free(buf);
	return ret;
}

static int name_comp(const void *a, const void *b)
{
	const struct user_summary_info *x = a, *y = b;
	char *n1 = x->ui->pw_name;
	char *n2 = y->ui->pw_name;

	if (!n1)
		return 1;
	if (!n2)
		return -1;
	return strcmp(n1, n2);
}

static int uid_comp(const void *a, const void *b)
{
	const struct user_summary_info *x = a, *y = b;
	return -NUM_COMPARE(x->ui->uid, y->ui->uid);
}

static int dir_count_comp(const void *a, const void *b)
{
	const struct user_summary_info *x = a, *y = b;
	return NUM_COMPARE(x->dirs, y->dirs);
}

static int file_count_comp(const void *a, const void *b)
{
	const struct user_summary_info *x = a, *y = b;
	return NUM_COMPARE(x->files, y->files);
}

static int size_comp(const void *a, const void *b)
{
	const struct user_summary_info *x = a, *y = b;
	return NUM_COMPARE(x->bytes, y->bytes);
}

static int count_admissible_users(__a_unused struct user_info *ui, void *data)
{
	struct user_summary_loop_data *usld = data;
	usld->num_admissible_users++;
	return 1;
}

static int print_user_summary(struct format_info *fi)
{
	int i, ret;
	int (*comp)(const void *a, const void *b);
	struct user_summary_loop_data usld = { .fi = fi};
	char *header = select_conf.header_given? select_conf.header_arg :
		"User summary\n";

	ret = output("%s", header);
	if (ret < 0)
		return ret;
	ret = for_each_admissible_user(count_admissible_users, &usld);
	if (ret < 0)
		return ret;
	if (usld.num_admissible_users == 0)
		return 1;
	usld.usis = adu_calloc(usld.num_admissible_users
		* sizeof(struct user_summary_info));
	usld.current = usld.usis;
	ret = for_each_admissible_user(compute_user_summary, &usld);
	if (ret < 0)
		goto out;
	switch (select_conf.user_summary_sort_arg) {
	case user_summary_sort_arg_name:
		comp = name_comp;
		break;
	case user_summary_sort_arg_uid:
		comp = uid_comp;
		break;
	case user_summary_sort_arg_dir_count:
		comp = dir_count_comp;
		break;
	case user_summary_sort_arg_file_count:
		comp = file_count_comp;
		break;
	case user_summary_sort_arg_size:
		comp = size_comp;
		break;
	default: /* this should never happen, but anyway */
		comp = size_comp;
		break;
	}
	qsort(usld.usis, usld.num_admissible_users,
		sizeof(struct user_summary_info), comp);
	for (i = 0; i < usld.num_admissible_users; i++) {
		if (select_conf.limit_arg >= 0 && i > select_conf.limit_arg)
			break;
		ret = print_user_summary_line(usld.usis + i, usld.fi);
		if (ret < 0)
			goto out;
	}
	ret = output("%s", select_conf.trailer_arg);
out:
	free(usld.usis);
	return ret;
}

static int user_list_loop_function(struct osl_row *row, void *data)
{
	struct user_list_info *uli = data;
	union atom_value values[] = {
		[ula_pw_name] = {.string_value = uli->ui->pw_name?
			uli->ui->pw_name : "?"},
		[ula_uid] = {.num_value = (long long unsigned)uli->ui->uid},
		[ula_files] = {.num_value = 0ULL},
		[ula_size] = {.num_value =  0ULL},
		[ula_dirname] = {.string_value = NULL}
	};
	uint64_t num;
	int ret;
	char *dirname = NULL, *buf;

	check_signals();
	ret = -E_LOOP_COMPLETE;
	if (!uli->count)
		goto err;

	ret = get_dir_name_of_user_row(row, uli->ui, &dirname);
	if (ret < 0)
		goto err;
	if (!dir_is_admissible(dirname, uli->preg, uli->inverse_matching)) {
		free(dirname);
		return 1;
	}
	values[ula_dirname].string_value = dirname;

	ret = get_num_user_files(row, uli->ui, &num);
	if (ret < 0)
		goto err;
	values[ula_files].num_value = num;

	ret = get_num_user_bytes(row, uli->ui, &num);
	if (ret < 0)
		goto err;
	values[ula_size].num_value = num;

	buf = format_items(uli->fi, values);
	free(dirname);
	dirname = NULL;
	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		goto err;
	uli->count--;
	return ret;
err:
	free(dirname);
	uli->ret = ret;
	uli->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int print_user_list(struct user_info *ui, void *data)
{
	struct user_list_format_info *ulfi = data;
	int ret;
	enum user_table_columns sort_column;
	struct user_list_info uli = {
		.ui = ui,
		.fi = ulfi->fi,
		.count = select_conf.limit_arg
	};
	union atom_value header_trailer_values[] = {
		[ulha_uid] = {.num_value = (long long unsigned)ui->uid},
		[ulha_pw_name] = {.string_value = ui->pw_name?
			ui->pw_name : "?"}
	};
	char *buf = format_items(ulfi->header_fi, header_trailer_values);

	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		return ret;
	if (select_conf.list_sort_arg == list_sort_arg_file_count)
		sort_column = UT_FILES;
	else
		sort_column = UT_BYTES;

	ret = compile_regex(&uli.preg, &uli.inverse_matching);
	if (ret < 0)
		return ret;
	ret = adu_loop_reverse(ui->table, sort_column, &uli,
		user_list_loop_function, &uli.ret, &uli.osl_errno);
	free_regex(uli.preg);
	if (ret < 0)
		return ret;
	buf = format_items(ulfi->trailer_fi, header_trailer_values);
	ret = output("%s", buf);
	free(buf);
	return ret;
}

static int print_user_lists(struct format_info *fi)
{
	struct user_list_format_info ulfi = {.fi = fi};
	char *header_fmt = select_conf.header_given?
		select_conf.header_arg : "uid %(uid)(%(pw_name)):\n";
	char *trailer_fmt = select_conf.trailer_arg;
	int ret = parse_format_string(header_fmt,
		user_list_header_trailer_atoms, &ulfi.header_fi);
	if (ret < 0)
		return ret;
	ret = parse_format_string(trailer_fmt,
		user_list_header_trailer_atoms, &ulfi.trailer_fi);
	if (ret < 0)
		return ret;
	ret = for_each_admissible_user(print_user_list, &ulfi);
	free_format_info(ulfi.header_fi);
	free_format_info(ulfi.trailer_fi);
	return ret;
}

static int global_list_loop_function(struct osl_row *row, void *data)
{
	struct global_list_info *gli = data;
	union atom_value values[] = {
		[gla_size] = {.num_value = 0ULL},
		[gla_files] = {.num_value =  0ULL},
		[gla_dirname] = {.string_value = NULL}
	};
	uint64_t num_files, num_bytes;
	char *dirname = NULL, *buf;
	int ret;

	check_signals();
	ret = -E_LOOP_COMPLETE;
	if (!gli->count)
		goto err;

	ret = get_dir_name_of_row(row, &dirname);
	if (ret < 0)
		goto err;
	if (!dir_is_admissible(dirname, gli->preg, gli->inverse_matching)) {
		free(dirname);
		return 1;
	}
	values[gla_dirname].string_value = dirname;

	ret = get_num_files_of_row(row, &num_files);
	if (ret < 0)
		goto err;
	values[gla_files].num_value = (long long unsigned)num_files;

	ret = get_num_bytes_of_row(row, &num_bytes);
	if (ret < 0)
		goto err;
	values[gla_size].num_value = (long long unsigned)num_bytes;

	buf = format_items(gli->fi, values);
	free(dirname);
	dirname = NULL;
	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		goto err;
	if (gli->count > 0)
		gli->count--;
	return ret;
err:
	free(dirname);
	gli->ret = ret;
	gli->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static int print_global_list(struct format_info *fi)
{
	int ret;
	enum dir_table_columns sort_column;
	struct global_list_info gli = {
		.fi = fi,
		.count = select_conf.limit_arg
	};
	char *header = select_conf.header_given?
		select_conf.header_arg : "Global list\n";

	ret = output("%s", header);
	if (ret < 0)
		return ret;
	if (select_conf.list_sort_arg == list_sort_arg_file_count)
		sort_column = DT_FILES;
	else
		sort_column = DT_BYTES;
	ret = compile_regex(&gli.preg, &gli.inverse_matching);
	if (ret < 0)
		return ret;
	ret = adu_loop_reverse(dir_table, sort_column, &gli,
		global_list_loop_function, &gli.ret, &gli.osl_errno);
	free_regex(gli.preg);
	if (ret < 0)
		return ret;
	return output("%s", select_conf.trailer_arg);
}

static int print_statistics(struct format_info *fi)
{
	switch (select_conf.select_mode_arg) {
	case select_mode_arg_global_list:
		return print_global_list(fi);
	case select_mode_arg_global_summary:
		return print_global_summary(fi);
	case select_mode_arg_user_list:
		return print_user_lists(fi);
	case select_mode_arg_user_summary:
		return print_user_summary(fi);
	default:
		ERROR_LOG("bad select mode\n");
		return -ERRNO_TO_ERROR(EINVAL);
	};
}

static int open_pipe(char *path)
{
	int p[2], ret;
	char **argv;

	ret = pipe(p);
	if (ret < 0)
		return ERRNO_TO_ERROR(errno);
	ret = fork();
	if (ret < 0)
		return ERRNO_TO_ERROR(errno);
	if (ret) { /* parent */
		DEBUG_LOG("created process %d\n", ret);
		close(p[0]);
		output_file = fdopen(p[1], "w");
		if (!output_file)
			return ERRNO_TO_ERROR(errno);
		return 1;
	}
	close(p[1]);
	if (p[0] != STDIN_FILENO)
		dup2(p[0], STDIN_FILENO);
	DEBUG_LOG("executing %s\n", path);
	split_args(path, &argv, " \t");
	execvp(argv[0], argv);
	ERROR_LOG("error executing %s: %s\n", path,
		adu_strerror(ERRNO_TO_ERROR(errno)));
	_exit(EXIT_FAILURE);
}

static int open_output_stream(void)
{
	char *p;
	int ret, flags = O_WRONLY | O_CREAT;

	if (!select_conf.output_given)
		goto use_stdout;
	p = select_conf.output_arg;
	switch (p[0]) {
	case '\0': /* empty string */
		goto bad_output_arg;
	case '-':
		if (!p[1]) /* "-" means stdout */
			goto use_stdout;
		/* string starting with a dash */
		flags |= O_EXCL;
		goto open_file;
	case '>':
		if (!p[1]) /* ">" is invalid */
			goto bad_output_arg;
		if (p[1] != '>') {
			p++;
			flags |= O_TRUNC;
			goto open_file;
		}
		/* string starting with ">>" */
		if (!p[2]) /* ">>" is invalid */
			goto bad_output_arg;
		flags |= O_APPEND;
		p += 2;
		goto open_file;
	case '|':
		if (!p[1]) /* "|" is invalid */
			goto bad_output_arg;
		p++;
		return open_pipe(p);
	default: /* args starts with no magic character */
		flags |= O_EXCL;
		goto open_file;
	}
use_stdout:
	output_file = stdout;
	return 1;
bad_output_arg:
	output_file = NULL;
	return -E_BAD_OUTPUT_ARG;
open_file:
	/*
	 * glibc's 'x' mode to fopen is not portable, so use open() and
	 * fdopen().
	 */
	ret = open(p, flags, 0644);
	if (ret < 0)
		return -ERRNO_TO_ERROR(errno);
	output_file = fdopen(ret, "w");
	if (!output_file)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

/**
 * Execute a select query.
 *
 * \param admissible_uids User IDs to take into account.
 * \param fi Format information.
 *
 * Called once in select mode or for each \a run command in interactive mode.
 *
 * Open the output stream and the dir table if not already open. For each
 * admissible uid, the user table is opened if necessary. After these
 * preparations, the output according to \a select_mode and \a fi is written to
 * the output stream.
 *
 * \return Standard.
 */
int run_select_query(struct uid_range *admissible_uids, struct format_info *fi)
{
	int ret = open_output_stream();

	if (ret < 0)
		goto out;
	ret = open_dir_table(0);
	if (ret < 0)
		goto out;
	check_signals();
	ret = open_admissible_user_tables(admissible_uids);
	if (ret < 0)
		goto out;
	check_signals();
	ret = print_statistics(fi);
out:
	if (output_file && output_file != stdout) {
		fclose(output_file);
		output_file = NULL;
	}
	return ret;
}

/** Default format string for global_list mode. */
#define GLOBAL_LIST_DFLT_FMT "%(size:r:8) %(files:r:8) %(dirname)\n"
/** Default format string for global_summary mode. */
#define GLOBAL_SUMMARY_DFLT_FMT "#directories: %(dirs), #files: %(files), size: %(size)\n"
/** Default format string for user_list mode. */
#define USER_LIST_DFLT_FMT "%(size:r:5) %(files:r:5) %(dirname)\n"
/** Default format string for user_summary mode. */
#define USER_SUMMARY_DFLT_FMT "%(pw_name:l:16) %(uid:r:6) %(dirs:r:5) %(files:r:5) %(size:r:5)\n"

static int setup_format_string(char *fmt, struct format_info **fi)
{
	struct atom *atoms;

	if (!fmt)
		INFO_LOG("using default format string\n");
	switch (select_conf.select_mode_arg) {
	case select_mode_arg_global_list:
		if (!fmt)
			fmt = GLOBAL_LIST_DFLT_FMT;
		atoms = global_list_atoms;
		break;
	case select_mode_arg_global_summary:
		if (!fmt)
			fmt = GLOBAL_SUMMARY_DFLT_FMT;
		atoms = global_summary_atoms;
		break;
	case select_mode_arg_user_list:
		if (!fmt)
			fmt = USER_LIST_DFLT_FMT;
		atoms = user_list_atoms;
		break;
	case select_mode_arg_user_summary:
		if (!fmt)
			fmt = USER_SUMMARY_DFLT_FMT;
		atoms = user_summary_atoms;
		break;
	default:
		ERROR_LOG("bad select mode\n");
		return -ERRNO_TO_ERROR(EINVAL);
	};
	INFO_LOG("format string: %s\n", fmt);
	return parse_format_string(fmt, atoms, fi);
}

/**
 * Parse a given format string.
 *
 * \param string The format string to parse.
 * \param params gengetopt parameters.
 * \param admissible_uids The array of admissible uid ranges.
 * \param fi The format info to be used with format_items().
 *
 * If \a string is not \p NULL, it is broken down into its components using
 * \ref create_argv() and the resulting argument vector is passed together with
 * \a params to gengetopt's command line parser. If --help or --detailed-help
 * was specified in \a string, the corresponding help text is printed and the
 * function returns zero.
 *
 * Otherwise, any --uid or --user options are parsed and transformed into an
 * array of admissible uids which is returned via \a admissible_uids.
 *
 * Finally, the format string given by --format (or the default format string
 * for the given select mode if no --format option was given in \a string) is
 * parsed as well resulting in a format_info structure which is returned via
 * \a fi. The caller uses the \a fi pointer later to format each output line.
 *
 * \return Negative on errors, zero if --help or --detailed-help was given,
 * positive otherwise.
 *
 * \sa format_items().
 */
int parse_select_options(char *string, struct select_cmdline_parser_params *params,
		struct uid_range **admissible_uids, struct format_info **fi)
{
	int ret, num_uid_ranges;
	const char **line;
	char *fmt = NULL;

	if (string) {
		int argc;
		char **argv;

		ret = create_argv(string, &argv);
		if (ret < 0)
			return ret;
		argc = ret;
		ret = select_cmdline_parser_ext(argc, argv, &select_conf, params);
		free_argv(argv);
		if (ret)
			return -E_SYNTAX;
		if (select_conf.help_given || select_conf.detailed_help_given)
			goto help;
		fmt = select_conf.format_arg;
	}
	ret = parse_uid_arg(select_conf.uid_arg, admissible_uids);
	if (ret < 0)
		return ret;
	num_uid_ranges = ret;
	ret = append_users(select_conf.user_arg, select_conf.user_given,
		admissible_uids, num_uid_ranges);
	if (ret < 0)
		return ret;
	return setup_format_string(fmt, fi);
help:
	line = select_conf.detailed_help_given?
		select_args_info_detailed_help : select_args_info_help;
	if (!output_file)
		output_file = stdout;
	for (; *line; line++) {
		ret = output("%s\n", *line);
		if (ret < 0)
			return ret;
	}
	return 0;
}

/**
 * Main function for select mode.
 *
 * \return Standard.
 */
int com_select(void)
{
	struct uid_range *admissible_uids = NULL;
	int ret;
	struct format_info *fi;
	struct select_cmdline_parser_params params = {
		.override = 1,
		.initialize = 1,
		.check_required = 1,
		.check_ambiguity = 1,
		.print_errors = 1
	};

	ret = parse_select_options(conf.select_options_arg, &params,
		&admissible_uids, &fi);
	if (ret > 0) {
		ret = read_uid_file();
		if (ret < 0)
			goto out;
		ret = run_select_query(admissible_uids, fi);
		free_format_info(fi);
	}
out:
	select_cmdline_parser_free(&select_conf);
	return ret;
}
